<title>{% block title %}{{ config['APP_NAME'] or 'Title' }}{% endblock %}</title>

<meta charset="utf-8"/>
<meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no"/>

{% block description %}
<meta name="description" content=""/>
{% endblock %}

<link rel="shortcut icon" type="image/x-icon" href="{{ url_for('web.static', filename='img/favicon.ico') }}"/>
<link rel="alternate" hreflang="en" href="" type="text/html"/>

{% block css %}
<link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/css/bootstrap.min.css"
      integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous"/>

<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css"
      integrity="sha384-wvfXpqpZZVQGK6TAh5PVlGOfQNHSoD2xbE+QkPxCAFlNEevoEH3Sl0sibVcOQVnN" crossorigin="anonymous"/>

<link rel="stylesheet" type="text/css" media="screen" href="{{ url_for('web.static', filename='css/main.css') }}"/>
{% endblock %}
