{% extends "layout/base.html" %}

{% block title %}
	{{ error }}
{% endblock %}

{% block jumbotron %}
{% endblock %}

{% block container %}
    <h1>{{ error }}</h1>
    <p>{{ message }}</p>
{% endblock %}
